/////////////////////////////////////////////////////////////////////////////////////
// File        : BP2595.h
// Function    : Application include for CPS model BP2595
// Author      : Jim Lamberson
// Copyright   : Copyright (C) 2021 CPS
/////////////////////////////////////////////////////////////////////////////////////

#ifndef _INC_APIBP2595_H_
#define	_INC_APIBP2595_H_

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __linux__ 
    #define APIFUNC
    #define APICC
#elif _WIN32
    #ifndef APIFUNC
    #define APIFUNC __declspec(dllimport)
    #endif
    #define APICC   __stdcall
#else
    // todo: 
    #define APIFUNC
    #define APICC
#endif

// Constants =======================================================

#define BP2595_MAX_DEVICES           8       // Max number of BP2595s connected to a host (limited by device selector rotary switch).

   
typedef enum BP2595_ERR {                       // Error codes; restricted to range [0:255] -----------------------
    // Common
    BP2595_ERR_OK                   = 0,            // Success
    // Device errors
    BP2595_ERR_OPCODE               = 1,            // Command not recognized.
    BP2595_ERR_ARGVAL               = 2,            // Illegal argument value.
    BP2595_ERR_MODE_LOCKOUT         = 3,            // Warning: Remote control prohibited in local mode; Local control prohibited in remote mode.
    BP2595_ERR_FAULT_LOCKOUT        = 4,            // Warning: Command failed due to fault condition (e.g., WriteSetpoint rejected while UVLO).
    // API errors
    BP2595_ERR_API_CLOSED           = 20,           // API is closed.
    BP2595_ERR_API_OPEN             = 21,           // Warning: API is already open (attempting to open while already open).
    BP2595_ERR_DEVICE_NOT_FOUND     = 30,           // Specified device is not detected.
    BP2595_ERR_CANT_OPEN_DEVICE     = 31,           // Failed to open device.
    BP2595_ERR_DEVICE_CLOSED        = 32,           // Device is closed.
    BP2595_ERR_DEVICE_OPEN          = 33,           // Warning: Device is already open.
    BP2595_ERR_DUPLICATE_ADDRESS    = 34,           // Two modules have the same address.
    BP2595_ERR_WAS_SHUTDOWN         = 40,           // USB connection was shutdown.
    BP2595_ERR_READ                 = 41,           // USB read error.
    BP2595_ERR_WRITE                = 42,           // USB write error.
    BP2595_ERR_OPENHIDLIB           = 43,           // Failed to open HIDLIB.
    BP2595_ERR_REPLY_MISMATCH       = 50,           // Reply doesn't match last issued command.
    BP2595_ERR_REPLY_VALUE          = 51,           // Illegal reply value.
    BP2595_ERR_TIMEOUT              = 60,           // Timed out waiting for message from device.
    BP2595_ERR_MALLOC               = 70,           // Memory allocation failed.
    BP2595_ERR_CREATE_RESOURCE      = 71,           // Failed to create mutex/semaphore/thread.
    BP2595_ERR_RESOURCE_BUSY        = 72,           // Resource is busy (locked by another thread).
    BP2595_ERR_SYSTEM               = 100,          // Unknown system error.
    BP2595_ERRCODE_MAX              = 0xFFFFFFFF    // ErrCode size = 32 bits
} BP2595_ERR;

// STATUS/EVENT SYSTEM ========================================================

// Flags common to both STATE_x and NOTIFY_x ------------------------------------

//  D = state data flag
//  E = event flag; if not also D then state flag is don't care (associated state info is elsewhere)
// Live faults (overlaps STATE, NOTIFY and STICKY)
#define SS_UVLO             (1 << 0)    // DE : state = undervoltage lockout (system input voltage too low). event = state changed.
#define SS_OVERTEMP         (1 << 1)    // DE : state = dev1 or dev2 over-temperature. event = state changed.
#define SS_FAIL             (1 << 2)    // DE : state = relay, dev1 or dev2 fault. event = state changed.
#define SS_COM_TOUT         (1 << 3)    // DE : state = communication watchdog timeout. event = state changed. NOTE: COPY OF STICKY_COM_TOUT ***********
// External signals (overlaps STATE and NOTIFY)
#define SS_REMOTE           (1 << 4)    // DE : state = remote/local control mode. event = state changed.
// Controls
#define SS_HVEN_SP          (1 << 5)    // DE : state = hv enable. event = state changed by remote command.
#define SS_HVEN_ASYNC       (1 << 6)    // -E : state = n/a. event = hv enable changed by local command or async event.
#define SS_VOUT_SP          (1 << 7)    // -E : state = n/a. event = hv setpoint changed by remote command.
#define SS_VOUT_ASYNC       (1 << 8)    // -E : state = n/a. event = hv setpoint changed by local command or async event.
// Spares
#define SS_HV_PRESENT       (1 << 9)    // DE : state = HV present on output. event = HV_present state changed.
// Spares
#define SS_SPARE_10         (1 << 10)   // -- : state = n/a. event = n/a.
#define SS_SPARE_11         (1 << 11)   // -- : state = n/a. event = n/a.
#define SS_SPARE_12         (1 << 12)   // -- : state = n/a. event = n/a.
#define SS_SPARE_13         (1 << 13)   // -- : state = n/a. event = n/a.
#define SS_SPARE_14         (1 << 14)   // -- : state = n/a. event = n/a.
// Meters
#define SS_DATAREPORT       (1 << 15)   // -E : state = n/a. event = data report arrived.

// Event log -------------------------------------------

// The event log maintains a record of events for diagnostics.
// Log changes are not reportable events.
// The log is polled via read command; resettable log flags are cleared upon read command.

// Logged status (note: overlaps fault<3:0>)
// Sticky faults  (overlaps STATE, NOTIFY and STICKY)
#define STICKY_UVLO         SS_UVLO         // Undervoltage lockout. SET UPON STATE_UVLO RISING EDGE. CLEAR UPON READ COMMAND.
#define STICKY_OVERTEMP     SS_OVERTEMP     // Dev1 or dev2 over-temperature. SET UPON STATE_OVERTEMP RISING EDGE. CLEAR UPON READ COMMAND.
#define STICKY_FAIL         SS_FAIL         // Relay, dev1 or dev2 fault. SET UPON STATE_FAIL RISING EDGE. CLEAR UPON READ COMMAND.
#define STICKY_COM_TOUT     SS_COM_TOUT     // Communication watchdog timeout.  SET UPON STATE_COM_TOUT RISING EDGE. CLEAR UPON READ COMMAND.
// Other (no overlaps)
#define STICKY_RESET        (1 << 4)        // CPU was reset. SET UPON DEVICE BOOT. CLEAR UPON READ COMMAND.
#define STICKY_ATTRDEF      (1 << 5)        // Valid personality stored in flash (not resettable). SET UPON BOOT IF PROGRAMMED. NEVER CLEARED.

// State flags -------------------------------------------

// Live faults (overlaps STATE, NOTIFY and STICKY)
#define STATE_UVLO          SS_UVLO         // Undervoltage lockout.
#define STATE_OVERTEMP      SS_OVERTEMP     // Multiplier over-temperature.
#define STATE_FAIL          SS_FAIL         // Relay, dev1 or dev2 fault.
#define STATE_COM_TOUT      SS_COM_TOUT     // Communication watchdog timeout. SET UPON COM TIMEOUT. CLEAR UPON STICKY_COM_TOUT CLEARED.
// External signals (overlaps STATE and NOTIFY)
#define STATE_REMOTE        SS_REMOTE       // Remote/local control mode.
// Controls (overlaps STATE and NOTIFY)
#define STATE_HVEN          SS_HVEN_SP      // HV enabled by client.
// Spares
#define STATE_SPARE_6       (1 << 6)
#define STATE_SPARE_7       (1 << 7)
#define STATE_SPARE_8       (1 << 8)
// Monitors (overlaps STATE and NOTIFY)
#define STATE_HV_PRESENT    SS_HV_PRESENT   // HV present on output.
// Spares
#define STATE_SPARE_10      (1 << 10)
#define STATE_SPARE_11      (1 << 11)
#define STATE_SPARE_12      (1 << 12)
#define STATE_SPARE_13      (1 << 13)
#define STATE_SPARE_14      (1 << 14)
#define STATE_SPARE_15      (1 << 15)

// Event notification flags --------------------------------

#define NOTIFY_NONE         0               // No pending notifications.
// Live faults  (overlaps STATE, NOTIFY and STICKY)
#define NOTIFY_UVLO         SS_UVLO         // Undervoltage lockout (system input voltage too low). event = state changed.
#define NOTIFY_OVERTEMP     SS_OVERTEMP     // Multiplier over-temperature. event = state changed.
#define NOTIFY_FAIL         SS_FAIL         // Relay, dev1 or dev2 fault. event = state changed.
#define NOTIFY_COM_TOUT     SS_COM_TOUT     // Communication watchdog timeout. event = state changed.
// External signals (overlaps STATE and NOTIFY)
#define NOTIFY_REMOTE       SS_REMOTE       // Remote/local control mode. event = state changed.
// Controls
#define NOTIFY_HVEN_CMD     SS_HVEN_SP      // HV enable changed by remote command.
#define NOTIFY_HVEN_ASYNC   SS_HVEN_ASYNC   // HV enable changed by local command or async event.
#define NOTIFY_SETPT_CMD    SS_VOUT_SP      // Setpoint changed by remote command.
#define NOTIFY_SETPT_ASYNC  SS_VOUT_ASYNC   // Setpoint changed by local command or async event.
// Monitors
#define NOTIFY_HV_PRESENT   SS_HV_PRESENT   // HV present (on output) changed.
// Spares
#define NOTIFY_SPARE_10     (1 << 10)
#define NOTIFY_SPARE_11     (1 << 11)
#define NOTIFY_SPARE_12     (1 << 12)
#define NOTIFY_SPARE_13     (1 << 13)
// Status log
#define NOTIFY_LOG_CHANGE   (1 << 14)       // Status log changed (at least one flag was set or cleared).
// Meters
#define NOTIFY_DATAREPORT   SS_DATAREPORT   // Data report arrived.

// sys.FailFlags
#define SYS_FAIL_RELAY_DRV  (1 << 0)        // relay failed to move to commanded position (fb signals detected but indicate incorrect position)
#define SYS_FAIL_RELAY_FB   (1 << 1)        // relay feedback signals failed (either both fb signals or neither fb signal asserted)
#define SYS_FAIL_POSDEV     (1 << 2)        // positive supply is malfunctioning -- see dev faults for details
#define SYS_FAIL_NEGDEV     (1 << 3)        // negative supply is malfunctioning -- see dev faults for details

// Flag groups --------------------------------------------

// Notification events
#define NOTIFY_FAULT        (NOTIFY_UVLO | NOTIFY_OVERTEMP | NOTIFY_FAIL | NOTIFY_COM_TOUT)
#define NOTIFY_EXTSIG       (NOTIFY_REMOTE)
#define NOTIFY_SETPT        (NOTIFY_SETPT_CMD | NOTIFY_SETPT_ASYNC)
#define NOTIFY_HVEN         (NOTIFY_HVEN_CMD  | NOTIFY_HVEN_ASYNC)
#define NOTIFY_CTRL         (NOTIFY_HVEN | NOTIFY_SETPT)
#define NOTIFY_LOG          (NOTIFY_LOG_CHANGE)
#define NOTIFY_ALL_STATUS   (NOTIFY_FAULT | NOTIFY_EXTSIG | NOTIFY_CTRL | NOTIFY_LOG)

// Faults
#define UCSD_FAULT          (SS_UVLO | SS_OVERTEMP)                                             // Faults that unconditionally invoke shutdown (vs. personality-based)
#define SS_FAULT            (SS_UVLO | SS_OVERTEMP | SS_FAIL | SS_COM_TOUT)                     // Any fault

// Event log
#define STICKY_FAULT        (STICKY_UVLO | STICKY_OVERTEMP | STICKY_FAIL | STICKY_COM_TOUT)     // Any sticky fault

#define STICKY_HOST_RESETTABLE   (STICKY_UVLO | STICKY_OVERTEMP | STICKY_FAIL | STICKY_RESET)   // Host-resettable sticky bits. Note: excludes com timeout.
#define STICKY_ALL_RESETTABLE    (STICKY_HOST_RESETTABLE | STICKY_RESET)                        // Host-resettable sticky bits. Note: excludes com timeout.

// Primitive types ===============================================================

typedef unsigned int    uint;

#include <limits.h>

#if (UCHAR_MAX == 0xff)
typedef unsigned char   u8;
#else
#error "UNSUPPORTED 8-BIT SIZE"
#endif

#if (USHRT_MAX == 0xFFFF)
typedef unsigned short  u16;
typedef signed short    s16;
#else
#error "UNSUPPORTED 16-BIT SIZE"
#endif

#if (UINT_MAX >= 0xFFFFFFFF)
typedef unsigned int    u32;
typedef signed int      s32;
#elif (ULONG_MAX >= 0xFFFFFFFFUL) 
typedef unsigned long   u32;
typedef signed long     s32;
#else
#error "UNSUPPORTED 32-BIT SIZE"
#endif

// Booleans are required to have value=1 when true
typedef u8      bool8;
typedef u16     bool16;
typedef uint    boole;

// Composite types ----------------------------------

typedef union BP2595_VERSION {              // VERSION NUMBER ----------
    u16 align16;
    struct {
        u8              Major;              //   Major version.
        u8              Minor;              //   Minor version.
        u16             Build;              //   Build number.
    };
} BP2595_VERSION;

typedef union BP2595_VERSION_INFO {         // DEVICE VERSION INFO ----------
    u16 align16;
    struct {
        BP2595_VERSION  DevFw;              //   Device firmware version number.
        char            PwbRev;             //   Device hardware revision letter.
    };
} BP2595_VERSION_INFO;

typedef struct DEV_METERS {                 // INTERNAL DEVICE NATIVE METERS ------------
    u16             AbsOutV;                //   Absolute HV output voltage (unsigned volts).
    u16             OutUA;                  //   HV output current (uA).
    u16             DeciDegC;               //   Temperature (0.1 C/bit).
} DEV_METERS;

typedef struct BP2595_METERS {              // API METER DATA -------------
    // Application meters
    s32             outV;                   //   System output volts
    u16             outUA;                  //   System output uA
    // Diagnostics meters
    u16             deciDegC;               //   System temperature (0.1 degrees C/bit)
    u16             inMV;                   //   System supply mV
    DEV_METERS      devMeters[2];           //   Positive/negative HV supply meters
} BP2595_METERS;

typedef struct STATUS_SYS {                 // STATUS SYSTEM ----------
    u16             Events;                 //   Notification events.
    u16             Live;                   //   Current status.
    u8              Log;                    //   Event log.
    u8              FailFlags;              //   Subsystem fail flags.
} STATUS_SYS;

typedef struct BP2595_STATUS {              // API STATUS INFO ----------
    s32             SetpointV;              //   HV setpoint in volts.
    STATUS_SYS      StatusSystem;           //   Status and notification system.
} BP2595_STATUS;

// API functions -----------------------------------

// Admin
APIFUNC BP2595_ERR APICC BP2595_OpenApi(int *devflags);
APIFUNC BP2595_ERR APICC BP2595_CloseApi(void);
APIFUNC BP2595_ERR APICC BP2595_OpenDevice(int devaddr, BP2595_STATUS *status);
APIFUNC BP2595_ERR APICC BP2595_CloseDevice(int devaddr);
APIFUNC BP2595_ERR APICC BP2595_SetComWatchdog(int devaddr, u16 interval);
APIFUNC BP2595_ERR APICC BP2595_GetComWatchdog(int devaddr, u16 *interval);
APIFUNC BP2595_ERR APICC BP2595_KeepAlive(int devaddr);
// Utility
APIFUNC BP2595_ERR APICC BP2595_GetApiVersion(BP2595_VERSION *ver);
APIFUNC BP2595_ERR APICC BP2595_GetDeviceVersions(int devaddr, BP2595_VERSION_INFO *ver);
APIFUNC const char * APICC BP2595_ErrString(int errcode);
// Control
APIFUNC BP2595_ERR APICC BP2595_ProgramSetpoint(int devaddr, s32 volts);
APIFUNC BP2595_ERR APICC BP2595_ProgramHvEnable(int devaddr, int enable);
// Status
APIFUNC BP2595_ERR APICC BP2595_GetStatusUpdate(int devaddr, BP2595_STATUS *status, int tmax);
APIFUNC BP2595_ERR APICC BP2595_ClearEventLog(int devaddr, u8 flags);
// Metering
APIFUNC BP2595_ERR APICC BP2595_SetMeterInterval(int devaddr, u16 interval);
APIFUNC BP2595_ERR APICC BP2595_GetMeterInterval(int devaddr, u16 *interval);
APIFUNC BP2595_ERR APICC BP2595_GetMeterUpdate(int devaddr, BP2595_METERS *meters, int tmax);

#ifdef __cplusplus
}
#endif

#endif // ifdef _INC_APIBP2595_H_
